package net.sf.sit;

import java.lang.reflect.Proxy;
import java.util.List;

import javax.validation.Validation;

import net.sf.seaf.exception.NotInitializedException;
import net.sf.sit.candidate.impl.MatchingSelector;
import net.sf.sit.candidate.impl.MethodNameMatcher;
import net.sf.sit.exception.impl.RethrowExceptionHandler;
import net.sf.sit.invoker.impl.DefaultInvoker;
import net.sf.sit.proxy.ServiceInvocationProxy;
import net.sf.sit.support.Configuration;
import net.sf.sit.validator.impl.AnnotationValidator;
import net.sf.sit.values.impl.DefaultConverter;
import net.sf.sit.values.impl.DozerMapFinder;
import net.sf.sit.values.impl.DozerMapper;
import net.sf.sit.values.impl.FillByLastMapper;
import net.sf.sit.values.impl.PadWithLastMapper;

import org.dozer.DozerBeanMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Configures SIT.
 * <p>
 * This is the default public entry point to SIT.
 * <p>
 * The simplest usage of SIT in Java is:
 * 
 * <pre>
 * SITConfigurer sit = new SITConfigurer();
 * sit.setFacadeType(Facade1.class);
 * sit.setProvider(new Provider1());
 * sit.init();
 * Facade1 facade = (Facade1) sit.getFacade();
 * // now invoke business methods on facade:
 * facade.businessMethod();
 * </pre>
 * 
 * A sample Spring configuration:
 * 
 * <pre>
 * &lt;bean id="mySit" class="net.sf.sit.SITConfigurer" init-method="init">
 *  &lt;property name="facadeType" value="some.facade.pkg.MyFacade" />
 *  &lt;property name="providers">
 *   &lt;list>
 *    &lt;bean class="some.provider.pkg.Provider1" />
 *    &lt;bean class="some.provider.pkg.Provider2" />
 *    &lt;bean class="some.provider.pkg.Provider3" />
 *   &lt;/list>
 *  &lt;/property>
 * &lt;/bean>
 * &lt;bean id="myFacade" factory-bean="mySit" factory-method="getFacade" />
 * </pre>
 */
public class SITConfigurer extends Configuration {

	Logger log = LoggerFactory.getLogger(SITConfigurer.class);
	private boolean inited = false;

	/**
	 * Default constructor.
	 */
	public SITConfigurer() {
	}

	/**
	 * Initialize SIT configurer.
	 * <p>
	 * Call this method before requesting the facade implementation (via
	 * getFacade()) to let SITConfigurer initialize itself.
	 * 
	 * @return This SITConfigurer to allow method chaining.
	 */
	public SITConfigurer init() {
		log.info("Initializing SITConfigurer with provided values: {}", this);
		boolean hasCustomDozerConfiguration = true;
		if (null == getValidator()) {
			setValidator(new AnnotationValidator(Validation
					.buildDefaultValidatorFactory().getValidator()));
		}
		if (null == getSelector()) {
			setSelector(new MatchingSelector(new MethodNameMatcher()));
		}
		if (0 == getDozerMappingFileNames().size()) {
			hasCustomDozerConfiguration = false;
			addToDozerMappingFileNames("sit-dozer-configuration.xml");
		}
		if (null == getDozer()) {
			setDozer(new DozerBeanMapper(
					(List<String>) getDozerMappingFileNames()));
		}
		if (hasCustomDozerConfiguration) {
			DozerMapFinder mapFinder = new DozerMapFinder(getDozer());
			if (null == getArgumentsMapper()) {
				setArgumentsMapper(new DozerMapper(getDozer(), mapFinder,
						new PadWithLastMapper()));
			}
			if (null == getResultsMapper()) {
				setResultsMapper(new DozerMapper(getDozer(), mapFinder,
						new FillByLastMapper()));
			}

		} else {
			if (null == getArgumentsMapper()) {
				setArgumentsMapper(new PadWithLastMapper());
			}
			if (null == getResultsMapper()) {
				setResultsMapper(new FillByLastMapper());
			}
		}
		if (null == getArgumentsConverter()) {
			setArgumentsConverter(new DefaultConverter(getDozer()));
		}
		if (null == getResultConverter()) {
			setResultConverter(new DefaultConverter(getDozer()));
		}
		if (null == getExceptionHandler()) {
			setExceptionHandler(new RethrowExceptionHandler());
		}
		if (null == getInvoker()) {
			setInvoker(new DefaultInvoker());
		}
		inited = true;
		log.info("Initialized SITConfigurer as: {}", this);
		return this;
	}

	/**
	 * Set providers to a single provider.
	 * <p>
	 * A helper method to avoid constructing collection for
	 * setProviders(Collection&lt;Object> providers) method when there is only
	 * one provider.
	 * 
	 * @param provider
	 *            Provider
	 */
	public void setProvider(Object provider) {
		setProviders(null);
		addToProviders(provider);
	}

	/**
	 * Return the facade. This creates the dynamic proxy for the facade type.
	 * 
	 * @return Facade object
	 */
	public Object getFacade() {
		if (!inited) {
			throw new NotInitializedException(
					"SITConfigurer is not initialized. "
							+ "Have you called the init() method?");
		}
		ServiceInvocationProxy proxy = new ServiceInvocationProxy(this);
		return Proxy.newProxyInstance(getFacadeType().getClassLoader(),
				new Class<?>[] { getFacadeType() }, proxy);
	}

}
