package net.sf.sit.candidate.impl;

import java.lang.reflect.Method;
import java.util.Map;

import net.sf.sit.candidate.Matcher;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Matcher that enables specifying overrides - explicit configuration of how to
 * match.
 * <p>
 * In case of no override applicable for the facade method, delegates to a
 * fall-back matcher.
 * <p>
 * Specify the overrides as a map, where key is the name of the facade method
 * and value is the signature of the candidate method consisting of the full
 * class name, separating dot '.', and the method name.
 * <p>
 * If the candidate method is declared by an implemented interface, the override
 * may be specified for the interface method.
 */
public class ConfigurableMatcher implements Matcher {

	private Matcher fallbackMatcher;
	private Map<String, String> overrides;
	private final Logger log = LoggerFactory
			.getLogger(ConfigurableMatcher.class);

	/**
	 * Default constructor.
	 */
	public ConfigurableMatcher() {
	}

	/**
	 * Full constructor.
	 * 
	 * @param overrides
	 *            Overrides
	 * @param fallbackMatcher
	 *            Fall-back matcher
	 */
	public ConfigurableMatcher(Map<String, String> overrides,
			Matcher fallbackMatcher) {
		this.fallbackMatcher = fallbackMatcher;
		this.overrides = overrides;
	}

	public boolean matches(Method facadeMethod, Method candidateMethod) {
		String override = null;
		if (null != getOverrides())
			override = getOverrides().get(facadeMethod.getName());
		if (null != override) {
			log.trace("Applying override {}", override);
			final String candidateSignature = getSignature(candidateMethod);
			boolean matches = candidateSignature.equals(override);
			matches = matchInterfaceMethods(matches, override, candidateMethod);
			log.trace("Candidate method signature {} matches: {}",
					candidateSignature, matches);
			return matches;
		}
		log.trace("Delegating to fallback matcher");
		return getFallbackMatcher().matches(facadeMethod, candidateMethod);
	}

	/**
	 * Get the simple method signature.
	 * 
	 * @param method
	 *            Method
	 * @return Simple signature
	 */
	private String getSignature(Method method) {
		StringBuilder sb = new StringBuilder();
		sb.append(method.getDeclaringClass().getName());
		sb.append('.');
		sb.append(method.getName());
		return sb.toString();
	}

	/**
	 * If candidate method on a class does not match directly, check methods of
	 * interfaces implemented by the class.
	 * <p>
	 * Enables the user to specify override for an interface method, not tying
	 * it to any particular implementation.
	 * 
	 * @param matches
	 *            Does candidate method on a class match?
	 * @param override
	 *            Override string
	 * @param candidateMethod
	 *            Candidate method
	 * @return Whether the override matches on an interface method
	 */
	private boolean matchInterfaceMethods(boolean matches, String override,
			Method candidateMethod) {
		Class<?> candidateType = candidateMethod.getDeclaringClass();
		if (!matches && !candidateType.isInterface()) {
			Class<?>[] interfaces = candidateType.getInterfaces();
			for (int i = 0; i < interfaces.length && !matches; i++) {
				Method[] methods = interfaces[i].getMethods();
				for (int m = 0; m < methods.length && !matches; m++) {
					Method method = methods[m];
					String signature = getSignature(method);
					boolean sameMethod = method.getName().equals(
							candidateMethod.getName());
					boolean matchesOverride = signature.equals(override);
					matches = sameMethod && matchesOverride;
				}
			}
		}
		return matches;
	}

	public Map<String, String> getOverrides() {
		return overrides;
	}

	/**
	 * Set overrides.
	 * 
	 * @param overrides
	 *            Overrides
	 */
	public void setOverrides(Map<String, String> overrides) {
		this.overrides = overrides;
	}

	public Matcher getFallbackMatcher() {
		return fallbackMatcher;
	}

	/**
	 * Set fall-back matcher.
	 * 
	 * @param fallbackMatcher
	 *            Fall-back matcher
	 */
	public void setFallbackMatcher(Matcher fallbackMatcher) {
		this.fallbackMatcher = fallbackMatcher;
	}

}
