package net.sf.sit.validator.impl;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Set;

import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.constraints.NotNull;

import net.sf.sit.validator.Validator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Validator driven by JSR-303 (javax.validation) annotations.
 */
public class AnnotationValidator implements Validator {

	private final javax.validation.Validator validator;
	private final Logger log = LoggerFactory
			.getLogger(AnnotationValidator.class);

	/**
	 * Default constructor.
	 * 
	 * @param validator
	 *            JSR-303 validator provider
	 */
	public AnnotationValidator(javax.validation.Validator validator) {
		this.validator = validator;
	}

	public void validate(Method facadeMethod, Object... arguments)
			throws ConstraintViolationException {
		if (null == arguments)
			return;
		Class<?> clazz = facadeMethod.getDeclaringClass();
		Annotation[][] argsAnns = facadeMethod.getParameterAnnotations();
		Set<ConstraintViolation<?>> violations = new HashSet<ConstraintViolation<?>>();
		for (int argIndex = 0; argIndex < arguments.length; argIndex++) {
			validateArgument(clazz, argsAnns[argIndex], violations, argIndex,
					arguments[argIndex]);
		}
		if (!violations.isEmpty()) {
			log.debug(
					"Raising validation exception, there are {} validation failures: {}",
					violations.size(), violations);
			throw new ConstraintViolationException("SIT validation failure.",
					violations);
		}
	}

	/**
	 * Validate an argument.
	 * 
	 * @param clazz
	 *            Facade type
	 * @param argsAnns
	 *            Argument's annotations
	 * @param violations
	 *            Summary constraint violations
	 * @param argIndex
	 *            Argument's index
	 * @param argument
	 *            Argument's value
	 */
	private void validateArgument(Class<?> clazz, Annotation[] argsAnns,
			Set<ConstraintViolation<?>> violations, int argIndex,
			Object argument) {
		if (null != argument) {
			log.debug("Validating method argument {}", argument);
			violations.addAll(validator.validate(argument));
		} else {
			if (isAnnotatedNotNull(argsAnns, null)) {
				log.debug("Mandatory method argument with index {} is null",
						argIndex);
				ConstraintViolation<?> cv = new NullArgumentConstraintViolation(
						clazz, argIndex);
				violations.add(cv);
			} else {
				log.debug("Optional method argument with index {} is null",
						argIndex);
			}
		}
	}

	/**
	 * Is argument annotated with a {@link NotNull} annotation (or related)?
	 * 
	 * @param annotations
	 *            Argument's annotations
	 * @param processed
	 *            Already processed annotations, to handle cyclic dependencies
	 *            between annotations
	 * @return True iff argument is mandatory
	 */
	private boolean isAnnotatedNotNull(Annotation[] annotations,
			HashSet<Annotation> processed) {
		if (null == processed)
			processed = new HashSet<Annotation>();
		for (int annotIndex = 0; annotIndex < annotations.length; annotIndex++) {
			Annotation annotation = annotations[annotIndex];
			if (!processed.contains(annotation)) {
				processed.add(annotation);
				log.trace("Checking annotation {} on NotNull.", annotation);
				if (annotation instanceof NotNull
						|| isAnnotatedNotNull(annotation.annotationType()
								.getAnnotations(), processed))
					return true;
			}
		}
		return false;
	}
}
