package net.sf.sit.values.impl;

import java.util.Collection;

import net.sf.seaf.util.convert.collection.ObjectConverter;
import net.sf.seaf.util.convert.collection.SequenceConverterUtils;
import net.sf.sit.values.Converter;

import org.dozer.Mapper;
import org.dozer.converters.DateFormatContainer;
import org.dozer.converters.PrimitiveOrWrapperConverter;
import org.dozer.util.MappingUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Default SIT converter.
 */
public class DefaultConverter implements Converter {

	private final Logger log = LoggerFactory.getLogger(DefaultConverter.class);
	private PrimitiveOrWrapperConverter simpleTypeConverter = new PrimitiveOrWrapperConverter();
	private DateFormatContainer dateFormatContainer = new DateFormatContainer(
			"yyyy-MM-dd");
	private Mapper dozer;
	private final SequenceConverterUtils sequenceConverterUtils = new SequenceConverterUtils();

	// constructors

	/**
	 * Default constructor.
	 */
	public DefaultConverter() {
	}

	/**
	 * Full constructor.
	 * 
	 * @param dozer
	 *            Dozer mapper
	 */
	public DefaultConverter(Mapper dozer) {
		this.dozer = dozer;
	}

	// interface method implementation

	public <TT> TT convert(Class<TT> targetType, Object sourceValue,
			Class<?> targetComponentTypeHint) {
		if (null == sourceValue) {
			log.debug("Converting null source value as null");
			return null;
		}
		Class<?> sourceType = sourceValue.getClass();
		log.debug("Converting source value [{}] of type {} to target type {}",
				new Object[] { sourceValue, sourceType, targetType });
		if (shouldConvertAsSimple(targetType, sourceType)) {
			return convertSimple(targetType, sourceValue);
		}
		if (targetType.isArray()
				|| Collection.class.isAssignableFrom(targetType)) {
			return convertSequence(targetType, sourceValue,
					targetComponentTypeHint);
		}
		return convertBean(targetType, sourceValue);
	}

	private <TT> boolean shouldConvertAsSimple(Class<TT> targetType,
			Class<?> sourceType) {
		boolean eitherIsSimple = MappingUtils.isPrimitiveOrWrapper(sourceType)
				|| MappingUtils.isPrimitiveOrWrapper(targetType);
		boolean noneIsEnum = (!Enum.class.isAssignableFrom(sourceType))
				&& (!Enum.class.isAssignableFrom(targetType));
		return eitherIsSimple && noneIsEnum;
	}

	@SuppressWarnings("unchecked")
	private <TT> TT convertSimple(Class<TT> targetType, Object sourceValue) {
		log.trace("Converting as simple value");
		TT converted = (TT) getSimpleTypeConverter().convert(sourceValue,
				targetType, getDateFormatContainer());
		log.debug("Converted as simple value: {}", converted);
		return converted;
	}

	@SuppressWarnings("unchecked")
	private <TT> TT convertSequence(Class<TT> targetType, Object sourceValue,
			Class<?> targetComponentTypeHint) {
		log.trace("Converting as sequence");
		if (null == targetComponentTypeHint)
			targetComponentTypeHint = targetType.getComponentType();
		final Class<?> targetComponentType = targetComponentTypeHint;
		TT converted = (TT) getSequenceConverterUtils().convert(sourceValue,
				targetType, targetComponentType,
				new ObjectConverter<Object, Object>() {
					public Object convert(Object source) {
						return getDozer().map(source, targetComponentType);
					}
				});
		log.debug("Converted as sequence: {}", converted);
		return converted;
	}

	private <TT> TT convertBean(Class<TT> targetType, Object sourceValue) {
		log.trace("Converting as bean");
		TT converted = getDozer().map(sourceValue, targetType);
		log.debug("Converted as bean: {}", converted);
		return converted;
	}

	// standard getters/setters

	public Mapper getDozer() {
		return dozer;
	}

	public final void setDozer(Mapper dozer) {
		this.dozer = dozer;
	}

	private PrimitiveOrWrapperConverter getSimpleTypeConverter() {
		return simpleTypeConverter;
	}

	public final void setSimpleTypeConverter(
			PrimitiveOrWrapperConverter simpleTypeConverter) {
		this.simpleTypeConverter = simpleTypeConverter;
	}

	private DateFormatContainer getDateFormatContainer() {
		return dateFormatContainer;
	}

	public final void setDateFormatContainer(
			DateFormatContainer dateFormatContainer) {
		this.dateFormatContainer = dateFormatContainer;
	}

	private SequenceConverterUtils getSequenceConverterUtils() {
		return sequenceConverterUtils;
	}

}
