package net.sf.sit.values.impl;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import net.sf.sit.support.Holder;

import org.dozer.DozerBeanMapper;
import org.dozer.Mapper;
import org.dozer.classmap.ClassMap;
import org.dozer.classmap.ClassMappings;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Finds out if a mapping configuration exists in Dozer.
 */
public class DozerMapFinder {

	private final Logger log = LoggerFactory.getLogger(DozerMapFinder.class);
	private final Mapper dozer;

	/**
	 * Full constructor.
	 * 
	 * @param dozer
	 *            Dozer mapper
	 */
	public DozerMapFinder(Mapper dozer) {
		this.dozer = dozer;
	}

	/**
	 * Return true iff a mapping configuration exists in Dozer.
	 * 
	 * @param mapId
	 *            Id of a mapping configuration
	 * @return True iff the mapping configuration exists
	 */
	public boolean existsMapId(String mapId) {
		try {
			DozerBeanMapper dbm = ((DozerBeanMapper) dozer);
			loadCustomMappings(dbm);
			ClassMappings customMappings = getCustomMappings(dbm);
			ClassMap mapping = customMappings.find(Holder.class, Holder.class,
					mapId);
			log.debug("Mapping {} found: {}", mapId, (null != mapping));
			return (null != mapping);
		} catch (Exception e) {
			log.debug("Cannot find mapping {}, caught exception: {}", mapId,
					e.getLocalizedMessage());
			return false;
		}
	}

	/**
	 * Make sure Dozer has loaded the mappings.
	 * 
	 * @param dbm
	 *            Dozer mapper
	 * @throws NoSuchMethodException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	private void loadCustomMappings(DozerBeanMapper dbm)
			throws NoSuchMethodException, IllegalAccessException,
			InvocationTargetException {
		Method method = DozerBeanMapper.class.getDeclaredMethod(
				"getMappingProcessor", new Class[] {});
		method.setAccessible(true);
		method.invoke(dbm, new Object[] {});
	}

	/**
	 * Retrieve mappings from Dozer.
	 * 
	 * @param dbm
	 *            Dozer mapper
	 * @return Dozer mappings
	 * @throws NoSuchFieldException
	 * @throws IllegalAccessException
	 */
	private ClassMappings getCustomMappings(DozerBeanMapper dbm)
			throws NoSuchFieldException, IllegalAccessException {
		Field field = DozerBeanMapper.class.getDeclaredField("customMappings");
		field.setAccessible(true);
		return (ClassMappings) field.get(dbm);
	}

}
