package net.sf.sit.values.impl;

import java.lang.reflect.Method;

import net.sf.sit.support.SourceValues;
import net.sf.sit.support.TargetValues;
import net.sf.sit.values.Mapper;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Maps a source array of values to a target array of values by delegating to a
 * Dozer {@link org.dozer.Mapper}.
 * <p>
 * First, finds a mapping configuration in Dozer by target method name (as in
 * {@code Method.getName()}, e.g. {@code updateCustomer}).
 * <p>
 * If no such mapping is found, finds mapping by the complete signature of the
 * method (as in {@code Method.toString()}, e.g.
 * {@code public com.foo.Customer com.foo.CustomerService.updateCustomer(com.foo.Customer)}
 * ).
 * <p>
 * If a Dozer mapping is found, passes it to the Dozer mapper to perform the
 * mapping.
 * <p>
 * If no Dozer mapping is found, delegates mapping to a fall-back mapper.
 * 
 * @see Method.getName()
 * @see Method.toString()
 */
public class DozerMapper implements Mapper {

	private final Logger log = LoggerFactory.getLogger(DozerMapper.class);

	private final org.dozer.Mapper dozer;
	private final Mapper fallbackMapper;
	private final DozerMapFinder mapFinder;

	/**
	 * Full constructor.
	 * 
	 * @param dozer
	 *            Dozer mapper
	 * @param mapFinder
	 *            Dozer map finder
	 * @param fallbackMapper
	 *            Fall-back mapper
	 */
	public DozerMapper(org.dozer.Mapper dozer, DozerMapFinder mapFinder,
			Mapper fallbackMapper) {
		this.dozer = dozer;
		this.fallbackMapper = fallbackMapper;
		this.mapFinder = mapFinder;
	}

	public Object[] map(Object[] values, int targetLength, Method targetMethod) {
		log.debug("Mapping source values {} to target length {}", values,
				targetLength);
		Object[] result;
		String mapId = findExistingMapId(targetMethod);
		if (null != mapId) {
			log.debug("Mapping with map-id {} exists, using it", mapId);
			SourceValues resultHolder = getDozer().map(
					new SourceValues(values), TargetValues.class, mapId);
			if (null == resultHolder.getArgs())
				result = new Object[targetLength];
			else
				result = resultHolder.getArgs().toArray();
		} else {
			log.debug("Mapping with map-id {} does not exist, "
					+ "using fallback mapper", mapId);
			result = getFallbackMapper()
					.map(values, targetLength, targetMethod);
		}
		log.debug("Mapped target values: {}", result);
		assert (result.length == targetLength) : "result.length="
				+ result.length + ", targetLength=" + targetLength;
		return result;
	}

	private String findExistingMapId(Method targetMethod) {
		String mapId = targetMethod.getName();
		if (existsMapId(mapId))
			return mapId;
		else {
			mapId = targetMethod.toString();
			if (existsMapId(mapId))
				return mapId;
			return null;
		}
	}

	private boolean existsMapId(String mapId) {
		log.debug("Looking for mapping with map-id {}", mapId);
		return getMapFinder().existsMapId(mapId);
	}

	private org.dozer.Mapper getDozer() {
		return dozer;
	}

	private Mapper getFallbackMapper() {
		return fallbackMapper;
	}

	private DozerMapFinder getMapFinder() {
		return mapFinder;
	}

}
